@@ -22,6 +22,7 @@ import (
 	"encoding/base64"
 	"fmt"
 	"sort"
+	"strings"
 	"testing"
 	"time"
 
@@ -30,6 +31,9 @@ import (
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
 	"sigs.k8s.io/controller-runtime/pkg/client"
+
+	"github.com/fluxcd/pkg/ssa/normalize"
+	"github.com/fluxcd/pkg/ssa/utils"
 )
 
 func TestApply(t *testing.T) {
@@ -58,7 +62,7 @@ func TestApply(t *testing.T) {
 		sort.Sort(SortableUnstructureds(objects))
 		var expected []string
 		for _, object := range objects {
-			expected = append(expected, FmtUnstructured(object))
+			expected = append(expected, utils.FmtUnstructured(object))
 		}
 
 		// verify the change set contains only created actions
@@ -153,6 +157,7 @@ func TestApply_Force(t *testing.T) {
 	secretName, secret := getFirstObject(objects, "Secret", id)
 	crbName, crb := getFirstObject(objects, "ClusterRoleBinding", id)
 	stName, st := getFirstObject(objects, "StorageClass", id)
+	_, svc := getFirstObject(objects, "Service", id)
 
 	// create objects
 	if _, err := manager.ApplyAllStaged(ctx, objects, DefaultApplyOptions()); err != nil {
@@ -173,7 +178,9 @@ func TestApply_Force(t *testing.T) {
 		}
 
 		// verify that the error message does not contain sensitive information
-		expectedErr := fmt.Sprintf("%s invalid, error: secret is immutable", FmtUnstructured(secret))
+		expectedErr := fmt.Sprintf(
+			"%s dry-run failed (Invalid): Secret \"%s\" is invalid: data: Forbidden: field is immutable when `immutable` is set",
+			utils.FmtUnstructured(secret), secret.GetName())
 		if diff := cmp.Diff(expectedErr, err.Error()); diff != "" {
 			t.Errorf("Mismatch from expected value (-want +got):\n%s", diff)
 		}
@@ -181,7 +188,10 @@ func TestApply_Force(t *testing.T) {
 
 	t.Run("force applies immutable secret", func(t *testing.T) {
 		// force apply
-		changeSet, err := manager.ApplyAllStaged(ctx, objects, ApplyOptions{Force: true, WaitTimeout: timeout})
+		opts := DefaultApplyOptions()
+		opts.Force = true
+
+		changeSet, err := manager.ApplyAllStaged(ctx, objects, opts)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -218,6 +228,57 @@ func TestApply_Force(t *testing.T) {
 		}
 	})
 
+	t.Run("force apply waits for finalizer", func(t *testing.T) {
+		secretClone := secret.DeepCopy()
+		{
+			secretWithFinalizer := secretClone.DeepCopy()
+
+			unstructured.SetNestedStringSlice(secretWithFinalizer.Object, []string{"fluxcd.io/demo-finalizer"}, "metadata", "finalizers")
+			if err := manager.client.Update(ctx, secretWithFinalizer); err != nil {
+				t.Fatal(err)
+			}
+		}
+
+		// remove finalizer after a delay, to ensure the controller handles a slow deletion
+		go func() {
+			time.Sleep(3 * time.Second)
+
+			secretWithoutFinalizer := secretClone.DeepCopy()
+			unstructured.SetNestedStringSlice(secretWithoutFinalizer.Object, []string{}, "metadata", "finalizers")
+			if err := manager.client.Update(ctx, secretWithoutFinalizer); err != nil {
+				panic(err)
+			}
+		}()
+
+		// update a value in the secret
+		err = unstructured.SetNestedField(secret.Object, "val-secret2", "stringData", "key")
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		// force apply
+		opts := DefaultApplyOptions()
+		opts.Force = true
+
+		changeSet, err := manager.ApplyAllStaged(ctx, objects, opts)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		// verify the secret was recreated
+		for _, entry := range changeSet.Entries {
+			if entry.Subject == secretName {
+				if diff := cmp.Diff(CreatedAction, entry.Action); diff != "" {
+					t.Errorf("Mismatch from expected value (-want +got):\n%s", diff)
+				}
+			} else {
+				if diff := cmp.Diff(UnchangedAction, entry.Action); diff != "" {
+					t.Errorf("Mismatch from expected value (-want +got):\n%s", diff)
+				}
+			}
+		}
+	})
+
 	t.Run("recreates immutable RBAC", func(t *testing.T) {
 		// update roleRef
 		err = unstructured.SetNestedField(crb.Object, "test", "roleRef", "name")
@@ -226,7 +287,10 @@ func TestApply_Force(t *testing.T) {
 		}
 
 		// force apply
-		changeSet, err := manager.ApplyAllStaged(ctx, objects, ApplyOptions{Force: true, WaitTimeout: timeout})
+		opts := DefaultApplyOptions()
+		opts.Force = true
+
+		changeSet, err := manager.ApplyAllStaged(ctx, objects, opts)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -261,7 +325,10 @@ func TestApply_Force(t *testing.T) {
 		}
 
 		// force apply selector
-		changeSet, err := manager.ApplyAllStaged(ctx, objects, ApplyOptions{ForceSelector: meta, WaitTimeout: timeout})
+		opts := DefaultApplyOptions()
+		opts.ForceSelector = meta
+
+		changeSet, err := manager.ApplyAllStaged(ctx, objects, opts)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -276,6 +343,28 @@ func TestApply_Force(t *testing.T) {
 			}
 		}
 	})
+
+	t.Run("force apply returns validation error", func(t *testing.T) {
+		// update to invalid yaml
+		err = unstructured.SetNestedField(svc.Object, "ClusterIPSS", "spec", "type")
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		// force apply objects
+		opts := DefaultApplyOptions()
+		opts.Force = true
+
+		_, err := manager.ApplyAllStaged(ctx, objects, opts)
+		if err == nil {
+			t.Fatal("expected validation error but got none")
+		}
+
+		// should return validation error
+		if !strings.Contains(err.Error(), "is invalid") {
+			t.Errorf("expected error to contain invalid msg but got: %s", err)
+		}
+	})
 }
 
 func TestApply_SetNativeKindsDefaults(t *testing.T) {
@@ -291,7 +380,7 @@ func TestApply_SetNativeKindsDefaults(t *testing.T) {
 
 	manager.SetOwnerLabels(objects, "app1", "default")
 
-	if err := SetNativeKindsDefaults(objects); err != nil {
+	if err := normalize.UnstructuredList(objects); err != nil {
 		t.Fatal(err)
 	}
 
@@ -302,6 +391,20 @@ func TestApply_SetNativeKindsDefaults(t *testing.T) {
 			t.Fatal(err)
 		}
 	})
+
+	// re-apply objects
+	changeSet, err := manager.ApplyAllStaged(ctx, objects, DefaultApplyOptions())
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// verify that the change set contains no changed objects
+	for _, entry := range changeSet.Entries {
+		if diff := cmp.Diff(UnchangedAction, entry.Action); diff != "" {
+			t.Errorf("Mismatch from expected value (-want +got):\n%s", diff)
+		}
+	}
+
 }
 
 func TestApply_NoOp(t *testing.T) {
@@ -317,7 +420,7 @@ func TestApply_NoOp(t *testing.T) {
 
 	manager.SetOwnerLabels(objects, "app1", "default")
 
-	if err := SetNativeKindsDefaults(objects); err != nil {
+	if err := normalize.UnstructuredList(objects); err != nil {
 		t.Fatal(err)
 	}
 
@@ -386,12 +489,11 @@ func TestApply_Exclusions(t *testing.T) {
 			t.Fatal(err)
 		}
 
+		opts := DefaultApplyOptions()
+		opts.ExclusionSelector = meta
+
 		// apply with exclusions
-		changeSet, err := manager.ApplyAll(ctx, objects, ApplyOptions{
-			Force:             false,
-			ExclusionSelector: meta,
-			WaitTimeout:       time.Second,
-		})
+		changeSet, err := manager.ApplyAll(ctx, objects, opts)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -411,7 +513,7 @@ func TestApply_Exclusions(t *testing.T) {
 		}
 
 		for _, entry := range changeSet.Entries {
-			if entry.Action != ConfiguredAction && entry.Subject == FmtUnstructured(configMap) {
+			if entry.Action != ConfiguredAction && entry.Subject == utils.FmtUnstructured(configMap) {
 				t.Errorf("Expected %s, got %s", ConfiguredAction, entry.Action)
 			}
 		}
@@ -436,6 +538,77 @@ func TestApply_Exclusions(t *testing.T) {
 	})
 }
 
+func TestApply_IfNotPresent(t *testing.T) {
+	timeout := 10 * time.Second
+	ctx, cancel := context.WithTimeout(context.Background(), timeout)
+	defer cancel()
+
+	meta := map[string]string{
+		"fluxcd.io/ssa": "IfNotPresent",
+	}
+
+	id := generateName("skip")
+	objects, err := readManifest("testdata/test1.yaml", id)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	_, ns := getFirstObject(objects, "Namespace", id)
+	_, configMap := getFirstObject(objects, "ConfigMap", id)
+	configMapClone := configMap.DeepCopy()
+	configMapClone.SetAnnotations(meta)
+
+	t.Run("creates objects", func(t *testing.T) {
+		// create objects
+		opts := DefaultApplyOptions()
+		opts.IfNotPresentSelector = meta
+
+		changeSet, err := manager.ApplyAllStaged(ctx, []*unstructured.Unstructured{ns, configMapClone}, opts)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		for _, entry := range changeSet.Entries {
+			if entry.Action != CreatedAction {
+				t.Errorf("Expected %s, got %s for %s", CreatedAction, entry.Action, entry.Subject)
+			}
+		}
+	})
+
+	t.Run("skips apply when annotated IfNotPresent", func(t *testing.T) {
+		opts := DefaultApplyOptions()
+		opts.IfNotPresentSelector = meta
+
+		changeSet, err := manager.Apply(ctx, configMapClone, opts)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		if changeSet.Action != SkippedAction {
+			t.Errorf("Diff found for %s", changeSet.String())
+		}
+	})
+
+	t.Run("resume apply when is annotated Override", func(t *testing.T) {
+		override := map[string]string{
+			"fluxcd.io/ssa": "Override",
+		}
+		configMapClone.SetAnnotations(override)
+
+		opts := DefaultApplyOptions()
+		opts.IfNotPresentSelector = meta
+
+		changeSet, err := manager.Apply(ctx, configMapClone, opts)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		if changeSet.Action != ConfiguredAction {
+			t.Errorf("Diff found for %s", changeSet.String())
+		}
+	})
+}
+
 func TestApply_Cleanup(t *testing.T) {
 	timeout := 10 * time.Second
 	ctx, cancel := context.WithTimeout(context.Background(), timeout)
@@ -469,7 +642,7 @@ func TestApply_Cleanup(t *testing.T) {
 
 	_, deployObject := getFirstObject(objects, "Deployment", id)
 
-	if err := SetNativeKindsDefaults(objects); err != nil {
+	if err = normalize.UnstructuredList(objects); err != nil {
 		t.Fatal(err)
 	}
 
@@ -727,7 +900,7 @@ func TestApply_Cleanup_Exclusions(t *testing.T) {
 
 	_, deployObject := getFirstObject(objects, "Deployment", id)
 
-	if err := SetNativeKindsDefaults(objects); err != nil {
+	if err = normalize.UnstructuredList(objects); err != nil {
 		t.Fatal(err)
 	}
 
@@ -775,3 +948,32 @@ func TestApply_Cleanup_Exclusions(t *testing.T) {
 		}
 	})
 }
+
+func TestApply_MissingNamespaceErr(t *testing.T) {
+	timeout := 10 * time.Second
+	ctx, cancel := context.WithTimeout(context.Background(), timeout)
+	defer cancel()
+
+	id := generateName("err")
+	objects, err := readManifest("testdata/test1.yaml", id)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	_, configMap := getFirstObject(objects, "ConfigMap", id)
+	unstructured.RemoveNestedField(configMap.Object, "metadata", "namespace")
+
+	_, err = manager.ApplyAllStaged(ctx, []*unstructured.Unstructured{configMap}, DefaultApplyOptions())
+	if !strings.Contains(err.Error(), "namespace not specified") {
+		t.Fatal("Expected namespace not specified error")
+	}
+}
+
+func containsItemString(s []string, e string) bool {
+	for _, a := range s {
+		if a == e {
+			return true
+		}
+	}
+	return false
+}
